#!/usr/bin/env python
#
# Copyright 2009 Marcelo Boveto Shima.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.    See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA    02110-1301, USA.
#
# Author: marceloshima@gmail.com (Marcelo Boveto Shima)
# TODO : copyright to canonical

import dbus
import dbus.service

import os
import logging
import sys
import gobject
import threading
import time
import subprocess

import TXSession
import tacix.TXUtils

logger = tacix.TXUtils.get_logger('txmanager', '/var/log/tacixmanager.log')

BUS_NAME="org.tacix"
INTERFACE_SERVER="org.tacix.Manager"
INTERFACE_SESSION="org.tacix.Sessions"
OBJECT_PATH="/org/tacix/Manager"

# Taken from jockey
class PermissionDeniedByPolicy(dbus.DBusException):
    _dbus_error_name = 'freenx.manager.PermissionDeniedByPolicy'

def WatchSession():
    while 1:
        time.sleep(2.0)

class TXManager( dbus.service.Object ):
    def __init__( self ):
        from dbus.mainloop.glib import DBusGMainLoop 
 	DBusGMainLoop(set_as_default=True)

        # Getting the system dbus
        self.bus = dbus.SystemBus( )
        
        bus_name = dbus.service.BusName(BUS_NAME,
                                        self.bus,
                                        do_not_queue=True)
        dbus.service.Object.__init__(self, bus_name, OBJECT_PATH)
        self.loop = gobject.MainLoop()

        # Getting the ConsoleKit object
        self.ck_manager_obj = self.bus.get_object ('org.freedesktop.ConsoleKit', '/org/freedesktop/ConsoleKit/Manager')
        self.ck_manager = dbus.Interface (self.ck_manager_obj, 'org.freedesktop.ConsoleKit.Manager')

        self.manager_sessions = {}

        self.dbus_info = dbus.Interface(self.bus.get_object('org.freedesktop.DBus',
                                        '/org/freedesktop/DBus/Bus', False), 'org.freedesktop.DBus')

        app_thread = threading.Thread(target=WatchSession, name='app_thread')
        app_thread.start()
 
    def run(self):
        try:
            self.loop.run()
        except KeyboardInterrupt:
            logger.debug("keyboardInterrupt")

    def receive_session_closed_signal(self, session_id):
        logger.debug("SessionClosed received: %s" % session_id)
        if '-' in session_id:
            session_id = session_id.split('-')[2]
        if session_id in self.manager_sessions:
            session = self.manager_sessions.pop(session_id)
            session.remove_from_connection()

    def get_session(self, session_id):
        logger.debug("Getting session %s" % session_id)
        if '-' in session_id:
            session_id = session_id.split('-')[2]
        if session_id not in self.manager_sessions:
            logger.debug ("Session %s not found on db" % session_id)
            logger.debug (self.manager_sessions.keys())
        return self.manager_sessions[session_id]

    def get_resumable_sessions(self, options):
        sessions = {}
        for session in self.manager_sessions.values():
            properties = session.get_properties()
            if properties['type'] == 'shadow' or properties['persistent'] == '0':
                continue
            if 'user' in options:
                if properties['user'] != options['user']:
                    continue
            if 'status' in options:
                if options['status'].find(properties['state']) == -1:
                    continue
            logger.debug("Adding session %s" % properties)
            sessions[session.get_property('display')] = properties
        return sessions

    def create_session(self, sender, guest_session=False, effective_user=None):
        nxsession = TXSession.TXSession (sender, self.bus, self, self.ck_manager_obj, self.ck_manager, self.dbus_info, logger, effective_user=effective_user, guest_session=guest_session)
        session_id = nxsession.nxoptions.get_property('uniqueid')
        self.manager_sessions[session_id] = nxsession
        logger.debug ("Session %s created" % session_id)

        object_path = nxsession.get_object_path()
        logger.debug (object_path.__repr__())
        session_obj = self.bus.get_object (BUS_NAME, object_path)
        session = dbus.Interface (session_obj, INTERFACE_SESSION)
        session.connect_to_signal("SessionClosed", self.receive_session_closed_signal, dbus_interface=INTERFACE_SESSION)
        return nxsession


    #### DBUS Methods/Signals
    @dbus.service.method( dbus_interface=INTERFACE_SERVER,
                       in_signature='s', out_signature='o', sender_keyword='sender', connection_keyword='conn' )
    def GetSession( self, session_id, sender, conn ):
        return self.get_session(session_id).get_object_path()

    @dbus.service.method( dbus_interface=INTERFACE_SERVER,
                       in_signature='', out_signature='ao', sender_keyword='sender', connection_keyword='conn' )
    def GetSessions( self, sender, conn ):
        logger.debug ("Getting sessions")
        ops = []
        for session in self.manager_sessions.values():
            ops.append(session.get_object_path())
        return ops

    @dbus.service.method( dbus_interface=INTERFACE_SERVER,
                       in_signature='a{ss}', out_signature='a{sa{ss}}', sender_keyword='sender', connection_keyword='conn' )
    def GetResumableSessions(self, options, sender=None, conn=None):
        return self.get_resumable_sessions(options)

    @dbus.service.method( dbus_interface=INTERFACE_SERVER,
                       in_signature='', out_signature='a{sa{ss}}', sender_keyword='sender', connection_keyword='conn' )
    def GetShadowableSessions(self, sender=None, conn=None):
        sessions = {}
        ck_sessions = self.ck_manager.GetSessions()
        for ck_op_session in ck_sessions:
            ck_session_obj = self.bus.get_object ('org.freedesktop.ConsoleKit', ck_op_session)
            ck_session = dbus.Interface (ck_session_obj, 'org.freedesktop.ConsoleKit.Session')
            display = ck_session.GetX11Display().replace(':', '')
            if display != '' and int(display) < 100:
                session_opts = {'display': display,
                    'type': 'console',
                    'uniqueid': '',
                    'cli_opts': '--------',
                    'depth': '',
                    'screen': 'x',
                    'available': 'Running',
                    'state': 'Running',
                    'session': 'X%s (Local)' % display}
                sessions[session_opts['display']] = session_opts

        for session_key in self.manager_sessions.keys():
            session = self.manager_sessions[session_key]
            sessions[session.get_property('display')] = session.get_properties()
        return sessions

    @dbus.service.method( dbus_interface=INTERFACE_SERVER,
                       in_signature='s', out_signature='o', sender_keyword='sender', connection_keyword='conn' )
    def CreateSession( self, user, sender, conn ):
        logger.debug ("Creating a new session")
        nxsession = self.create_session(sender, effective_user=user)
        object_path = nxsession.get_object_path()

        return object_path

    @dbus.service.method( dbus_interface=INTERFACE_SERVER,
                       in_signature='', out_signature='s', sender_keyword='sender', connection_keyword='conn' )
    def CreateGuestSession( self, sender, conn ):
        logger.debug ("Setting a guest session up")
        nxsession = self.create_session(sender, guest_session=True)
        object_path = nxsession.get_object_path()

        return object_path

    # Taken from jockey 
    def _check_polkit_privilege(self, sender, conn, privilege):
        '''Verify that sender has a given PolicyKit privilege.

        sender is the sender's (private) D-BUS name, such as ":1:42"
        (sender_keyword in @dbus.service.methods). conn is
        the dbus.Connection object (connection_keyword in
        @dbus.service.methods). privilege is the PolicyKit privilege string.

        This method returns if the caller is privileged, and otherwise throws a
        PermissionDeniedByPolicy exception.
        '''
        if sender is None and conn is None:
            # called locally, not through D-BUS
            return False

        # get peer PID
        if self.dbus_info is None:
            self.dbus_info = dbus.Interface(conn.get_object('org.freedesktop.DBus',
                '/org/freedesktop/DBus/Bus', False), 'org.freedesktop.DBus')
        pid = self.dbus_info.GetConnectionUnixProcessID(sender)

        # query PolicyKit
        if self.polkit is None:
            self.polkit = dbus.Interface(dbus.SystemBus().get_object(
                'org.freedesktop.PolicyKit', '/', False),
                'org.freedesktop.PolicyKit')
        try:
            res = self.polkit.IsProcessAuthorized(privilege, pid, True)
        except dbus.DBusException, e:
            if e._dbus_error_name == 'org.freedesktop.DBus.Error.ServiceUnknown':
                # polkitd timed out, connect again
                self.polkit = None
                return self._check_polkit_privilege(sender, conn, privilege)
            else:
                raise

        if res != 'yes':
            logger.debug('_check_polkit_privilege: sender %s on connection %s pid %i requested %s: %s' % (
                sender, conn, pid, privilege, res))
            raise PermissionDeniedByPolicy(privilege + ' ' + res)
        return True

