#!/usr/bin/env python
#
# Copyright 2009 Marcelo Boveto Shima.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.    See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA    02110-1301, USA.
#
# Author: marceloshima@gmail.com (Marcelo Boveto Shima)

import dbus
import dbus.service
import os, os.path
import subprocess
import pwd
import random

from tacix import TXUtils
from tacix.nx.NXOptions import NXOptions

INTERFACE_SESSION="org.tacix.Sessions"
SESSION_TYPE="TXManager"
SESSION_PATH="/org/tacix/Sessions/"

class TXSession( dbus.service.Object ):
    def __init__( self, sender, bus, manager, ck_manager_obj, ck_manager, dbus_info, logger, guest_session=False, effective_user=None):
        self.logger = logger

        # Getting the system dbus
        self.bus = bus
        
        # Helper objects
        self.manager = manager
        self.ck_manager_obj = ck_manager_obj
        self.ck_manager = ck_manager
        self.ck_session = None
        self.dbus_info = dbus_info
        self.agent_pid = None

        self.sender = sender
        self.session_process = None

        self.guest_session = guest_session
        if self.guest_session:
            guest_name = self.create_guest_user()
            if guest_name is not None:
                effective_user = guest_name

        if effective_user is None:
            return

        self.effective_pwd = pwd.getpwnam(effective_user)

        # Properties
        self.ck_cookie = None
        self.nxoptions = NXOptions(str(self.effective_pwd.pw_uid))
        self.printers = {}

        ses_id = self.nxoptions.get_id()
        self.object_path = dbus.ObjectPath(SESSION_PATH + ses_id)
        dbus.service.Object.__init__(self, bus, self.object_path)

        self.logger.debug( "Watch for conn: %s" % sender )
        self.node_signal = self.bus.watch_name_owner(sender, self.connection_node_closed_signal)


    def get_object_path(self):
        return self.object_path

    def get_property(self, property):
        value = self.nxoptions.get_property(property)
        self.logger.debug( "Getting property %s with value %s" % (property, value))
        return value

    def get_properties(self):
        return self.nxoptions.get_properties()

    def run_dialog(self, authority_file, display):
        ret = TXUtils.run_program("/usr/lib/tacix/tacix-dialog", authority_file, display)
        self.logger.debug('tacix-dialog returned %d' % ret)
        if os.path.lexists(authority_file):
            os.remove(authority_file)
        return ret

    def ask_cookie(self):
        self.logger.debug("Asking cookie from session %s" % self.nxoptions.get_id())
        tmp_authority_file = "/tmp/.nx-authority-%s" % str(random.random())
        cookie = self.nxoptions.get_cookie()
        display = ":" + self.nxoptions.get_display()

        if not TXUtils.add_cookie_to_xauthority(tmp_authority_file, display, cookie):
            return None

        if self.run_dialog(tmp_authority_file, display) == 1:
            return cookie
        self.logger.debug("Authorization = %s" % str(retcode))
        return None

    def ask_console_authority(self):
        display = ":" + self.nxoptions.get_property('shadowdisplay')

        ck_sessions = self.ck_manager.GetSessions()
        local_ck_session = None
        for ck_op_session in ck_sessions:
            ck_session_obj = self.bus.get_object ('org.freedesktop.ConsoleKit', ck_op_session)
            ck_session = dbus.Interface (ck_session_obj, 'org.freedesktop.ConsoleKit.Session')
            ck_display = ck_session.GetX11Display()
            self.logger.debug('Comparing %s and %s' % (display, ck_display))
            if display == ck_display:
                self.logger.debug('ConsoleKit session found')
                local_ck_session = ck_session
                break

        if local_ck_session == None:
            self.logger.debug('ConsoleKit session not found')
            return False

        pwd_user = pwd.getpwuid(local_ck_session.GetUser())
        console_authority_file = '%s/.Xauthority' % pwd_user.pw_dir
        tmp_authority_file = "/tmp/.nx-authority-%s" % str(random.random())
        self.logger.debug('Asking to merge display %s on %s into %s' %\
                (display, console_authority_file, tmp_authority_file))
        if not TXUtils.merge_cookie_from_xauthority(console_authority_file, tmp_authority_file, display):
            self.logger.debug('Failed to merge from console authority')
            self.logger.debug('console = %s tmp = %s display = %s ' %\
                    (console_authority_file, tmp_authority_file, display))
            return false

        authority_file = self.nxoptions.get_property('authority_file')
        if self.run_dialog(tmp_authority_file, display) == 1:
            return TXUtils.merge_cookie_from_xauthority(console_authority_file, authority_file, display)
        return False

    def run_as_user(self):
        if self.effective_pwd is None:
            return False
        pwd_user = self.effective_pwd

        new_env = {}
        new_env['USER'] = pwd_user.pw_name
        new_env['USERNAME'] = pwd_user.pw_name
        new_env['LOGNAME'] = pwd_user.pw_name
        new_env['HOME'] = pwd_user.pw_dir
        new_env['NXSESSIONID'] = self.nxoptions.options['id']
        for env_var in ['LANG', 'LANGUAGE', 'PATH']:
            if env_var in os.environ:
                new_env[env_var] = os.environ[env_var]

        cmd = "/usr/lib/tacix/tacix-session"
        self.logger.debug("Running tacix-session")
        self.session_process = subprocess.Popen(cmd, env=new_env, close_fds=True, stdin=None, stdout=None, stderr=None, preexec_fn=lambda: os.setuid(pwd_user.pw_uid))
        return True

    def connection_node_closed_signal(self, sender):
        if sender == '':
            self.logger.debug("Connection with node closed")
            self.node_signal.remove()
            if self.nxoptions.get_state() != 'suspended':
                self.CloseSession()

    def connection_session_closed_signal(self, sender):
        if sender == '':
            self.logger.debug("Connection closed, session %s exited " % self.nxoptions.get_id())
            self.CloseSession()

    def create_guest_user(self):
        process = subprocess.Popen("/usr/lib/tacix/guest-session-setup.sh", stdout=subprocess.PIPE, stderr=None, stdin=None)
        user = None
        while True:
            line = process.stdout.readline()
            if len(line) == 0:
                break
            user = line
        ret = process.wait()
        if ret == 0 and user is not None:
            user = user.strip()
            self.guest_user = user
            return user
        return None

    def cleanup_guest_user(self):
        ret = process = subprocess.call(["/usr/lib/tacix/guest-session-cleanup.sh", self.guest_user])
        if ret == 0:
            return True
        return False


    # Signals
    @dbus.service.signal( INTERFACE_SESSION )
    def StateChanged (self, state):
        self.logger.debug( "StateChanged emitted, state = %s" % state )
        pass

    @dbus.service.signal( INTERFACE_SESSION )
    def PrinterAdded (self, printer_name):
        self.logger.debug( "Printer %s added" % printer_name )
        pass

    @dbus.service.signal( INTERFACE_SESSION )
    def PrinterSolicited (self, printer_name):
        self.logger.debug( "Printer %s solicited" % printer_name )
        pass

    @dbus.service.signal( INTERFACE_SESSION )
    def InfoReceived (self, property):
        self.logger.debug( "InfoReceived emitted, property = %s" % property )
        pass

    @dbus.service.signal( INTERFACE_SESSION )
    def SessionClosed (self, message):
        self.logger.debug( "SessionClosed emitted" )
        pass

    @dbus.service.signal( INTERFACE_SESSION )
    def SessionResumed (self, message):
        self.logger.debug( "SessionResumed emitted" )
        pass

    @dbus.service.signal( INTERFACE_SESSION )
    def SessionSuspended (self, message):
        self.logger.debug( "SessionSupended emitted" )
        pass

    # Methods 
    @dbus.service.method( dbus_interface=INTERFACE_SESSION,
                       in_signature='s', out_signature='', sender_keyword='sender', connection_keyword='conn' )
    def SendInfo( self, info, sender, conn ):
        self.logger.debug("Info %s received" % info)
        self.InfoReceived( info )

    @dbus.service.method( dbus_interface=INTERFACE_SESSION,
                       in_signature='', out_signature='b', sender_keyword='sender', connection_keyword='conn' )
    def AskShadowPermission(self, sender, conn):
        self.logger.debug('Asking for shadow permission')
        if self.nxoptions.get_property('type') != 'shadow':
            self.logger.debug('Session type is not shadow')
            return False
        shadowed_id = self.nxoptions.get_property('shadowed_id')
        self.logger.debug('shadow id is %s' % shadowed_id)
        if shadowed_id is not None:
            session = self.manager.get_session(shadowed_id)
            if session is None:
                self.logger.debug('Shadowed session not found')
                return False
            cookie = session.ask_cookie()
            if cookie is None:
                return False
            authority_file = self.nxoptions.get_property('authority_file')
            return TXUtils.add_cookie_to_xauthority(authority_file, ":" + self.nxoptions.get_property('shadowdisplay'), cookie)
        else:
            self.logger.debug('Shadowing a local session')
            return self.ask_console_authority()

    @dbus.service.method( dbus_interface=INTERFACE_SESSION,
                       in_signature='ss', out_signature='', sender_keyword='sender', connection_keyword='conn' )
    def SetProperty( self, property, value, sender, conn ):
        self.logger.debug( "Setting property %s with value %s" % (property, value) )
        self.nxoptions.set_property(property, value)

    @dbus.service.method( dbus_interface=INTERFACE_SESSION,
                       in_signature='a{ss}', out_signature='', sender_keyword='sender', connection_keyword='conn' )
    def SetProperties( self, properties, sender, conn ):
        self.logger.debug( "Setting properties %s" % properties )
        for prop in properties:
            self.SetProperty(prop, properties[prop], sender, conn)

    @dbus.service.method( dbus_interface=INTERFACE_SESSION,
                       in_signature='s', out_signature='s', sender_keyword='sender', connection_keyword='conn' )
    def GetProperty( self, property, sender, conn ):
        if str(self.bus.get_unix_user(sender)) == self.nxoptions.get_property('uid'):
            return self.nxoptions.get_property_secure(property)
        return self.nxoptions.get_property(property)

    @dbus.service.method( dbus_interface=INTERFACE_SESSION,
                       in_signature='', out_signature='a{ss}', sender_keyword='sender', connection_keyword='conn' )
    def GetProperties( self, sender, conn ):
        if str(self.bus.get_unix_user(sender)) == self.nxoptions.get_property('uid'):
            properties = self.nxoptions.get_properties_secure()
        else:
            properties = self.nxoptions.get_properties()
        self.logger.debug( "Getting properties %s" % properties )
        return properties

    @dbus.service.method( dbus_interface=INTERFACE_SESSION,
                       in_signature='s', out_signature='', sender_keyword='sender', connection_keyword='conn' )
    def SetState( self, state, sender, conn ):
        if self.nxoptions.set_state(state) is None:
            self.logger.debug("State didn't changed. Old is %s and new is %s" % (old_state, state))
            return
        self.StateChanged( state )
        self.logger.debug( "State changed to %s" % state)

    @dbus.service.method( dbus_interface=INTERFACE_SESSION,
                       in_signature='', out_signature='s', sender_keyword='sender', connection_keyword='conn' )
    def GetState( self, sender, conn ):
        return self.nxoptions.get_property('state')

    @dbus.service.method( dbus_interface=INTERFACE_SESSION,
                       in_signature='', out_signature='', sender_keyword='sender', connection_keyword='conn' )
    def CloseSession( self, sender=None, conn=None ):
        if self.ck_session is not None:
            self.ck_manager.CloseSession (self.ck_cookie)
            self.ck_session = None

        if self.session_process is not None:
            self.session_process.wait()

        if self.guest_session:
            self.cleanup_guest_user()

        self.SessionClosed(self.nxoptions.get_property('id'))
        pass

    @dbus.service.method( dbus_interface=INTERFACE_SESSION,
                       in_signature='', out_signature='b', sender_keyword='sender', connection_keyword='conn' )
    def Run( self, sender, conn ):
        # The sender that owns the session can run a guest session.
        if self.guest_session and sender == self.sender:
            self.run_as_user()
            return True
        # Otherwise only the user can run his session
        if self.effective_pwd.pw_uid == self.bus.get_unix_user(sender):
            self.run_as_user()
            return True
        return False

    @dbus.service.method( dbus_interface=INTERFACE_SESSION,
                       in_signature='a{ss}', out_signature='', sender_keyword='sender', connection_keyword='conn' )
    def AskPrinter( self, properties, sender, conn ):
        self.logger.debug("Asking for printer with parameters %s" % properties)
        if 'printer' in properties:
            self.printers[properties['printer']] = properties
            self.PrinterSolicited(properties['printer'])

    @dbus.service.method( dbus_interface=INTERFACE_SESSION,
                       in_signature='s', out_signature='a{ss}', sender_keyword='sender', connection_keyword='conn' )
    def GetPrinter( self, printer_name, sender, conn ):
        if printer_name not in self.printers:
            return {}
        return self.printers[printer_name]

    @dbus.service.method( dbus_interface=INTERFACE_SESSION,
                       in_signature='s', out_signature='s', sender_keyword='sender', connection_keyword='conn' )
    def OpenSession( self, remote_host, sender, conn ):
        uid = self.bus.get_unix_user(sender)
        if str(uid) != self.nxoptions.get_property('uid'):
            # The session cannot be openned by another user
            self.CloseSession()
            return ''

        self.logger.debug ("Openning the session")
        #self._check_polkit_privilege(sender, conn, INTERFACE_SERVER)
        self.bus.watch_name_owner(sender, self.connection_session_closed_signal)
        self.nxoptions.set_property('txsession_pid', str(self.dbus_info.GetConnectionUnixProcessID(sender)))
        try:

            # Defining the session attributes
            params = dbus.Array ([], signature = "(sv)")
            params.append (("unix-user", dbus.Int32 (uid, variant_level=1)))
            params.append (("session-type", dbus.String (SESSION_TYPE, variant_level=1)))
            params.append (("x11-display", dbus.String (":" + self.nxoptions.get_property('display'), variant_level=1)))
            params.append (("x11-display-device", dbus.String ("", variant_level=1)))
            params.append (("display-device", dbus.String ("", variant_level=1)))
            params.append (("remote-host-name", dbus.String (remote_host, variant_level=1)))
            params.append (("is-local", dbus.Boolean (True, variant_level=1)))

            # Create the ConsoleKit session
            self.ck_cookie = self.ck_manager.OpenSessionWithParameters (params)
            self.logger.debug (self.ck_cookie)

            # Getting the ConsoleKit session
            current_session = self.ck_manager.GetSessionForCookie (self.ck_cookie)
            session = current_session

            session_obj = self.bus.get_object ('org.freedesktop.ConsoleKit', current_session)
            session = dbus.Interface (session_obj, 'org.freedesktop.ConsoleKit.Session')
            self.ck_session = session

            # Setting the session as active
            properties = dbus.Interface (session_obj, 'org.freedesktop.DBus.Properties')
            properties.Set ("org.freedesktop.DBus.Properties", "active", dbus.Boolean (True, variant_level=1))

            return self.ck_cookie

        except dbus.DBusException, e:
            # Dbus error problably you don't have the dbus rule installed or your launcher is not suid nx
            # Open session without the parameters
            log.error ("Failed to create a CK session using parameters")
    
            error_string = str(e)
            log.error ("Error: " + error_string)
            return ""


