#!/usr/bin/env python

import os
import subprocess
import paramiko
import sys
import re
import select
import socket
import time
import SocketServer
import threading
import gobject
gobject.threads_init()

import tacix.TXUtils
from tacix.freenx.FreeNXShell import FreeNXShell

GENERIC_LINE = re.compile(r'^NX> (?P<protnr>\d+) (?P<protcmd>.*)')
PROPERTIES = { '700': ('Session id', 'id'),
        '705': ('Session display', 'display'),
        '703': ('Session type', 'type'),
        '701': ('Proxy cookie', 'proxy_cookie'),
        '702': ('Proxy IP', 'proxyip'),
        '706': ('Agent cookie', 'cookie'),
        '704': ('Session cache', 'type'),
        '707': ('SSL tunneling', 'encryption')}

STATUS = ['1009', '710']
IGNORED = ['716', '1002', '105']

DEFAULT_OPTIONS = {'link':'modem',
        'backingstore': '1',
        'encryption': '0',
        'cache': '16M',
        'images': '64M',
        'shmem': '1',
        'shpix': '1',
        'strict': '0',
        'composite': '1',
        'media': '0',
        'session': 'localhost',
        'type': 'unix-gnome',
        'geometry': '1440x850',
        'client': 'linux',
        'keyboard': 'pc101/us',
        'screeninfo': '1440x850x24+render'}

class TXClient(FreeNXShell):
    def __init__(self, host='localhost', port=22):
        self.options = {}
        self.options.update(DEFAULT_OPTIONS)
        self.options['geometry'] = self.get_geometry()
        
        self.txpath = None
        self.host = host
        self.port = port
        self.logger = None

        self.resumablesessions = None
        self.stage = None
        self.channels = None
        self.nxproxy_proc = None

        self.connect_callback = None
        self.started_signal_callback = None
        self.disconnect_callback = None

    # Options manipulations functions
    def write_options(self):
        opts = ['nx/nx']

        for opt_name in ("keyboard", "id", "clipboard", "geometry",
                "option", "shadow", "shadowmode", "streaming", "defer", "tile",
                "R", "render", "fullscreen", "shpix", "shmem", "encryption",
                "resize", "backingstore", "menu", "shadowuid", "clients",
                "client", "copysize", "cookie"
                # Server only properties
                # "cache", "images", "type", "link"
                ):
            if opt_name in self.options:
                opts.append('%s=%s' % (opt_name, self.options[opt_name]))

        open( self.txpath.get_filepath( "options" ), "w").write("%s:%s\n" % (",".join(opts), self.options['display']))

    def get_geometry(self):
        xprop = subprocess.Popen(["/usr/bin/xprop", "-root", "_NET_WORKAREA"], stdout=subprocess.PIPE)
        line = xprop.stdout.read(1024)
        xprop.wait()
        line = line.strip()
        xprops = line.split('=')[1].split(',')
        width = xprops[2].strip()
        height = str(int(xprops[3].strip()) -25)
        return width + 'x' + height

    # Protocol handler
    def handle_line(self, line):
        if line is None: 
            return True
        line = line.strip()
        if line == '':
            return True

        self.debug(line)
        match = GENERIC_LINE.search(line)
        if match is None:
            if self.stage is not None and self.stage == 'listsession':
                return self.add_running_session(line)
            if line == 'NX> 105':
                return True
            self.debug('Could not parse line %s' % line)
            return True
        protnr = match.group('protnr')
        protcmd = match.group('protcmd')
        self.debug(protnr)
        self.debug(protcmd)
        if protnr in IGNORED:
            return True
        if protnr in STATUS:
            return True
        if protnr in PROPERTIES:
            prop, key = PROPERTIES[protnr]
            value = protcmd.split(':')[1].strip()
            self.options[key] = value
            self.debug('Added value %s to key %s' % (value, key))
            return True
        handler_name = '_%s_handler' % protnr
        try:
            handler_method = getattr(self, handler_name)
            self.debug('handler %s found' % handler_name)
            return handler_method(line)
        except AttributeError, e:
            self.debug('Unhandled nx command %r (%s)' % (protnr, e))

        return True

    def _1000_handler(self, line):
        self.stdin.write('listsession --user="%(user)s" --type="%(type)s"' % self.options)
        self.stdin.write('\n')
        self.stdin.flush()
        return True

    def _127_handler(self, line):
        self.stage = 'listsession'
        return True

    def _148_handler(self, line):
        self.stage = None
        if self.resumablesessions is not None:
            for session in self.resumablesessions:
                if session['state'] != 'terminated':
                    self.options['id'] = self.resumablesessions[0]['uniqueid']
        cmd = []
        for property in self.options.keys():
            cmd.append('--%s="%s"' % (property, self.options[property]))
        send = None
        if 'id' in self.options:
            send = 'restoresession %s' % ' '.join(cmd)
        else:
            send = 'startsession %s' % ' '.join(cmd)
        self.stdin.write(send + '\n')
        self.debug(send)
        self.stdin.flush()
        return True

    def _1001_handler(self, line):
        return self._bye_handler(line)

    def _999_handler(self, line):
        return self._bye_handler(line)

    def _1006_handler(self, line):
        self.stdin.write('bye')
        self.stdin.write('\n')
        self.stdin.flush()
        self.run_proxy()
        return True

    def _bye_handler(self, line):
        if self.channels is not None:
            for chan in self.channels:
                chan.close()
        if self.nxproxy_proc is not None:
            self.nxproxy_proc.wait()
        if self.disconnect_callback is not None:
            self.disconnect_callback()
        return False

    def add_running_session(self, line):
        options = {}
        options.update({
                  'display': line[:7].strip(),
                  'type': line[8:24].strip(),
                  'uniqueid': line[25:57].strip(),
                  'cli_opts': line[58:66].strip(),
                  'depth': line[67:72].strip(),
                  'screen': line[73:87].strip(),
                  'state': line[88:99].strip(),
                  'session': line[100:].strip(),
                  })
        if options['display'] == 'Display' or options['display'] == '-------':
            return True
        self.debug(options)
        if self.resumablesessions is None:
            self.resumablesessions = [options]
        else:
            self.resumablesessions.append(options)
        return True

    # Actions
    def connect(self, user, password):
        self.options['user'] = user
        self.client = paramiko.SSHClient()
        self.client.load_system_host_keys()
        self.client.set_missing_host_key_policy(paramiko.AutoAddPolicy())

        try:
            self.client.connect(self.host, port=self.port, username=user, password=password, allow_agent=False, look_for_keys=False, timeout=10)
        except paramiko.AuthenticationException, e:
            return False
        command = '/usr/lib/tacix/tacix-freenx-node --slave'
        self.chan = self.client.get_transport().open_session()
        self.chan.exec_command(command)
        self.chan.settimeout(0.0)
        self.stdin = self.chan.makefile('wb', -1)
        self.stdout = self.chan.makefile('rb', -1)
        return True

    def debug(self, log):
        if self.logger is not None:
            self.logger.debug(log)
        else:
            print(log)

    def run(self, connect_callback=None, disconnect_callback=None, started_signal_callback=None):
        self.started_signal_callback = started_signal_callback
        self.connect_callback = connect_callback
        self.disconnect_callback = disconnect_callback
        gobject.io_add_watch(self.chan, gobject.IO_IN | gobject.IO_HUP, self.read_line)

    def wait_loop(self):
        self.loop = gobject.MainLoop()
        self.loop.run()
        self.debug('loop exited')

    def read_line(self, source, condition):
        if condition == gobject.IO_HUP:
            return False
        ret = False
        try:
            x = self.chan.recv(1024)
            self.debug(x)
            ret = self._receive(x, self.handle_line)
        except Exception, e:
            pass
        return ret

    def run_proxy(self):
        self.txpath = tacix.TXUtils.TXPath( self.options['id'], create=True, client=True )
        self.write_options( )
        os.environ['NX_DISPLAY'] = 'nx/nx,options=%s:%s' % (self.txpath.get_filepath('options'), self.options['display'])
        command = []
        self.chan.settimeout(None)
        display = self.options['display']
        #display = str(int(self.options['display']) + 1)

        command.append('/usr/bin/nxproxy')
        command.append('-S')
        if self.options['encryption'] == '1':
            display_port = int(display) + 4000
            aux = 0
            while True:
                try:
                    self.thread_tunnel = threading.Thread(target=self.tunnel_remote_port, args=(display_port, display_port))
                    self.thread_tunnel.start()
                    break
                except Exception, e:
                    self.debug('%s' % e)
                    aux = aux + 1
                    if aux == 3:
                        sys.exit(1)

            command.append('%s:%s' % ('localhost', display))
        else:
            command.append('%s:%s' % (self.host, display))
        if self.started_signal_callback is not None:
            self.started_signal_callback()
        self.nxproxy_proc = subprocess.Popen(command, close_fds=True)
        self.thread_tunnel.join()

    def pulseaudio(self):
        '''
           pulseaudio --use-pid-file=false --file pulseaudio.conf -n
           pulseaudio.conf:
           load-module module-esound-protocol-tcp auth-anonymous=1
           load-module module-native-protocol-tcp auth-anonymous=1

           server-side:
           PULSE_SERVER=localhost:4000
        '''
        return

    def tunnel_remote_port(self, remote_port, local_port):
        HOST = ''                 # Symbolic name meaning all available interfaces
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.bind((HOST, local_port))
        s.listen(1)
        conn, addr = s.accept()
        print 'Connected by', addr
        chan = None
        try:
            chan = self.client.get_transport().open_channel('direct-tcpip',
                                                   ('127.0.0.1', remote_port),
                                                   conn.getpeername())
        except Exception, e:
            raise Exception('Incoming request to %s:%d failed: %s' %
                    ('127.0.0.1', remote_port, repr(e)))
        if chan is None:
            raise Exception('Incoming request to %s:%d was rejected by the SSH server.' %
                    ('127.0.0.1', remote_port))

        gobject.io_add_watch(chan, gobject.IO_IN | gobject.IO_HUP, self.forward_input, conn)
        gobject.io_add_watch(conn, gobject.IO_IN | gobject.IO_HUP, self.forward_input, chan)
        self.channels = [chan, conn]

    def forward_input(self, source, condition, dest):
        if condition == gobject.IO_HUP:
            return False
        source.settimeout(0.0)
        data = source.recv(4096)
        if len(data) == 0:
            return False
        dest.send(data)
        source.settimeout(None)
        return True

