#!/usr/bin/env python
#
# Copyright 2009 Marcelo Boveto Shima.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.    See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA    02110-1301, USA.
#
# Author: marceloshima@gmail.com (Marcelo Boveto Shima)

import sys
import paramiko
import select
import traceback
import subprocess
import socket
import time
import posix
import re

from tacix.freenx.FreeNXServerProtocol import FreeNXServerProtocol
from tacix.freenx.FreeNXNodeProtocol import FreeNXNodeProtocol
from tacix.freenx.FreeNXShell import *

#GENERIC_INPUT = re.compile(r'^NX (?P<protnr>\d+) (?P<protcmd>.*)$ (?P<protact>.*)$')

info_lines = {
        'display': re.compile(r"^NX> 705 Session display: (?P<display>\d+)"),
        'tunneling': re.compile(r"^NX> 707 SSL tunneling: (?P<tunneling>\d+)"),
        'sessionid': re.compile(r"^NX> 700 Session id: (?P<sessionid>.*)$")
    }


class FreeNXServer(FreeNXServerProtocol, FreeNXNodeProtocol, FreeNXShell):
    def __init__( self, logger=None ):
        FreeNXShell.__init__(self, logger=logger)

        self.nodeprocess = None
        self.hostname = 'localhost'
        self.flags = None

        self.tunneling = 0
        self.display = 0
        self.sessionid = None

    def start_node(self, user, password):
        self.client = paramiko.SSHClient()
        self.client.load_system_host_keys()
        self.client.set_missing_host_key_policy(paramiko.AutoAddPolicy())

        self.client.connect(self.hostname, username=user, password=password, allow_agent=False, look_for_keys=False)
        command = '/usr/lib/tacix/tacix-freenx-node --slave'
        self.chan = self.client._transport.open_session()
        self.chan.exec_command(command)
        self.chan.settimeout(0.1)
        self.stdin = self.chan.makefile('wb', -1)

        return True

    def receive_chan_node(self):
        try:
            read = self.chan.recv(1024)
            return self._receive(read, self.read_node)
        except socket.timeout:
            self.logger.debug('socket timeout')
        return True

    def receive_stdin(self):
        #allowed_cmds = ['login', 'hello', 'set', 'quit', 'bye']
        allowed_cmds = None
        return self._receive_stdin(self.handle_line, allowed_cmds=allowed_cmds)

    def receive_stdin_guest(self):
        #allowed_cmds = ['startsession']
        allowed_cmds = None
        return self._receive_stdin(self.client_read_line_guest, allowed_cmds=allowed_cmds)

    def receive_logged_stdin(self):
        return self._receive_stdin(self.client_logged_read_line)

    def logged_loop(self):
        self._loop({sys.stdin:self.receive_stdin, self.chan:self.receive_chan_node}, message='waiting logged')

    def loop_guest(self):
        self._loop({sys.stdin:self.receive_stdin_guest}, message='waiting guest')

    def read_node(self, line):
        if self.logger is not None:
            self.logger.debug('node :' + line)

        if line.startswith('NX> 1000 NXNODE'):
            return True
        if line.startswith('NX> 716 Slave mode'):
            return True

        for info, rx in info_lines.iteritems():
            match = rx.search(line)
            if match:
                if info == 'display':
                    self.display = int(match.group('display'))
                    self.logger.debug('Display matched: %d' % self.display)
                    break
                elif info == 'sessionid':
                    self.sessionid = match.group('sessionid')
                    self.logger.debug('Session id matched: %s' % self.sessionid)
                    break
                elif info == 'tunneling':
                    self.tunneling = int(match.group('tunneling'))
                    self.logger.debug('Tunneling matched: %s' % self.tunneling)
                    break

        sys.stdout.write(line + '\n')
        sys.stdout.flush()
        return True

    def client_logged_read_line(self, line):
        if self.logger is not None:
            self.logger.debug('client_logged :' + line)

        return self.handle_line(line)

    def client_read_line_guest(self, line):
        if self.logger is not None:
            self.logger.debug('guest: ' + line)

        return self.handle_line(line, prefix='guest')

    def forward_line_node(self, line):
        sys.stdout.write(line + '\n')
        sys.stdout.flush()
        self.stdin.write(line + "\n")
        self.stdin.flush()
        return True

    def _guest_listsession_handler(self, line):
        opts = self.opt_read(line)
        self.say_listsessions_header(opts)
        self.say_listsessions_footer(opts)
        return True

    def _listsession_handler(self, line):
        return self.forward_line_node(line)

    def _addprinter_handler(self, line):
        return self.forward_line_node(line)

    def _startsession_handler(self, line):
        return self.forward_line_node(line)

    def _restoresession_handler(self, line):
        return self.forward_line_node(line)

    def _attachsession_handler(self, line):
        return self.forward_line_node(line)

    def _terminate_handler(self, line):
        return self.forward_line_node(line)

    def _quit_handler(self, line):
        return False

    def _bye_handler(self, line):
        self.say_server_bye(line)
        if self.tunneling == 1:
            self.logger.debug('Starting tunneled session')
            # If encrypted
            if True:
                # Netcat version
                self.revert_blocking(sys.stdin.fileno())
                self.logger.debug('netcat started')
                subprocess.call(['/bin/netcat', '127.0.0.1', str(self.display + 4000)], stdin=sys.stdin, stdout=sys.stdout, stderr=sys.stderr)
                self.logger.debug('netcat exited')
                self.set_non_blocking(sys.stdin.fileno())
            else:
                # Sockets version, not working
                s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                s.connect(("127.0.0.1", 5001))
                fd = s.fileno()
                #posix.close(sys.stdin.fileno())
                #posix.close(sys.stdout.fileno())
                posix.dup2(fd, sys.stdin.fileno())
                posix.dup2(fd, sys.stdout.fileno())
                #posix.close(fd)
                #posix.dup2(sys.stdin.fileno(), fd)
                #posix.dup2(sys.stdout.fileno(), fd)
                while True:
                    time.sleep(0.1)
        else:
            self.logger.debug('Starting non tunneled session')
            # TODO not tunneled server will never close, a proper fix is needed
            while True:
                time.sleep(0.1)

        return False

    def _hello_handler(self, line):
        self.say_server_accepted(line)
        return True

    def _set_handler(self, line):
        sys.stdout.write(line + '\n')
        sys.stdout.write('NX> 105 ')
        sys.stdout.flush()
        return True

    def _login_handler(self, line):
        self.say_server_user(line)
        self.current_state = 'user'
        user = None
        password = None

        r, w, e = select.select([sys.stdin], [], [])
        if sys.stdin in r:
            if self.logger is not None:
                self.logger.debug('reading user')
            user = sys.stdin.read().strip()
            self.logger.debug('client:' + user)

            if user == "NX guest user":
                self.logger.debug('Guest user authentication')
                self.say_server_welcome(self.hostname, user)
                self.loop_guest()
                return False

            self.say_server_password(line)

            # getpass is not working probably nxclient don't send the char needed
            # we need another way to don't prompt the password
            #self.revert_blocking(sys.stdin.fileno())
            #password = getpass.getpass(prompt="", stream=None)
            #self.set_non_blocking(sys.stdin.fileno())

            #import termios
            #fd = sys.stdin.fileno()
            #old = termios.tcgetattr(fd)
            #new = termios.tcgetattr(fd)
            #new[3] = new[3] & ~termios.ECHO          # lflags
            #try:
            #    if self.logger is not None:
            #        self.logger.debug('reading password')
            #    termios.tcsetattr(fd, termios.TCSADRAIN, new)
            #    r, w, e = select.select([sys.stdin], [], [])
            #    if sys.stdin in r:
            #        password = sys.stdin.read().strip()
            #finally:
            #    termios.tcsetattr(fd, termios.TCSADRAIN, old)

        r, w, e = select.select([sys.stdin], [], [])
        if sys.stdin in r:
            if self.logger is not None:
                self.logger.debug('reading password')
            password = sys.stdin.read().strip()

            if self.start_node(user, password):
                self.say_server_welcome(self.hostname, user)
                self.logged_loop()
                self.logger.debug('Closing the connection')
                self.client.close()
                self.logger.debug('Connection closed')
            else:
                self.say_server_login_failed()
        return False


    def run( self ):
        self.say_server_hello()
        sys.stdout.flush()
        try:
            FreeNXShell.run(self)
        except Exception, e:
            self.say_abort(e)

