#!/usr/bin/python
#
# Copyright 2007 Google Inc.
# Copyright 2009 Marcelo Boveto Shima.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.    See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA    02110-1301, USA.
#
# Author: diamond@google.com (Stephen Shirley)
#         marceloshima@gmail.com (Marcelo Boveto Shima)

import os
import signal
import subprocess
import sys
import shutil

from tacix import TXUtils

class NXAgent:
    def __init__( self, nxsession, options ):
        self.options = options
        self.txpath = nxsession.txpath
        self.txsession = nxsession.txsession
        self.logger = self.txpath.get_logger("nxagent")
        self.process = None

    # Utilities methods
    def send_signal(self, num):
        try:
            self.logger.info('Send signal %d to agent %d' % (num, self.process.pid))
            os.kill(self.process.pid, num)
            return True
        except OSError, e:
            return False


    # Options manipulations functions
    def write_options(self):
        opts = ['nx/nx']

        for opt_name in ("keyboard", "id", "clipboard", "geometry",
                "option", "shadow", "shadowmode", "streaming", "defer", "tile",
                "R", "render", "fullscreen", "shpix", "shmem", "link",
                "media", "cups", "samba", "cache", "images",
                "resize", "backingstore", "menu", "shadowuid", "clients",
                "client", "copysize", "cookie" ):
            if opt_name in self.options:
                opts.append('%s=%s' % (opt_name, self.options[opt_name]))
        opts.append("type=" + self.options['type'].replace('unix-', ''))

        open( self.txpath.get_filepath( "options" ), "w").write("%s:%s\n" % (",".join(opts), self.options['display']))


    # Action methods
    def terminate(self):
        self.send_signal(signal.SIGTERM)
        self.cleanup()

    def run(self):
        self.logger.info("Starting session %s\n" % self.options['id'])
        self.logger.debug("Exporting DISPLAY for nxagent:")
        display_tmp = 'nx/nx,options=%s:%s' % (self.txpath.get_filepath( "options" ), self.options['display'])
        os.environ['DISPLAY'] = display_tmp

        if not self.setup_xauth():
            self.txsession.SetState('denied')
            return False
        self.start_agent()
        os.environ['DISPLAY'] = ':%s' % self.options['display']
        return True

    def start_agent(self):
        self.logger.info("Starting nxagent\n")
        self.write_options()
        cmd = []
        cmd.append('/usr/bin/nxagent')
        if self.options['type'] == 'shadow':
            cmd.append('-S')
            cmd.append('-shadow')
            #cmd.append('%(shadowhost)s:%(shadowdisplay)s' % self.options)
            cmd.append(':%(shadowdisplay)s' % self.options)
            cmd.append('-shadowmode')
            cmd.append('1') #Interactive showdowing [1|0]
        else:
            if self.options['rootless'] == '1': cmd.append('-R')
            else: cmd.append('-D')

            if self.options['persistent'] == '1': cmd.append('-persistent')
            else: cmd.append('-nopersistent')

        cmd.append('-name')
        cmd.append('"TaciX - %s@%s:%s - %s" ' % \
                     (self.options['user'], os.uname()[1], self.options['display'], self.options['session']))
        cmd.append(':%s' % self.options['display'])

        self.logger.debug("Agent cmd: %s\n" % cmd)

        self.process = subprocess.Popen(cmd,
                stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        self.logger.info("Spawned nxagent pid %d\n" % self.process.pid)

    def setup_xauth(self):
        """Setup Xauthority file using session cookie."""
        authority_file = self.txpath.get_filepath('authority')
        os.environ['XAUTHORITY'] = authority_file
        self.txsession.SetProperty('authority_file', authority_file)

        if not TXUtils.add_cookie_to_xauthority(authority_file, 'localhost:%(display)s' % self.options, self.options['cookie']):
            return False
        if not TXUtils.add_cookie_to_xauthority(authority_file, ':%(display)s' % self.options, self.options['cookie']):
            return False

        if self.options['type'] == 'shadow':
            os.environ['SHADOW_XAUTHORITY'] = self.txpath.get_filepath('authority')
            # Ask authorization
            if not self.txsession.AskShadowPermission():
                return False
        return True

    def cleanup_xauth(self):
        """Setup Xauthority file using session cookie."""
        authority_file = self.txpath.get_filepath('authority')
        TXUtils.remove_display_from_xauthority(authority_file, 'localhost:%(display)s' % self.options)
        TXUtils.remove_display_from_xauthority(authority_file, ':%(display)s' % self.options)

    def cleanup(self, status=None, keep_log=True):
        """Cleanup after various things (X11 locks etc)."""

        self.logger.info('Cleaning up session\n')
        for proc in 'watchdog', 'agent', 'app':
            pid_name = "%s_pid" % proc
            if pid_name in self.options:
                try:
                    os.kill(int(self.options[pid_name]), signal.SIGTERM)

	            # sleep then kill (force)
                    #os.kill(self.options[pid_name], signal.SIGKILL)
                except OSError, (errno, strerror):
                    self.logger.warn("Cleanup handler got error %d when "
                            "killing %s: %s\n" % (errno, proc, strerror))
                else:
                    self.logger.warn("Cleanup handler sent TERM to %s\n" % proc)

        for f in ['/tmp/.X%(display)s-lock' % self.options,
                  '/tmp/.nX%(display)s-lock' % self.options,
                  '/tmp/.X11-unix/X%(display)s' % self.options]:
            if os.path.exists(f):
                try:
                    os.remove(f)
                except OSError, e:
                    self.logger.error("Error removing file. %s\n" % e)

        #self.cleanup_xauth()

        # Move if failed or terminated if defined
        if not keep_log:
            shutil.rmtree(self.txpath.get_dir())
        elif status == 'terminated':
            shutil.move(self.txpath.get_dir(), self.txpath.get_terminated_dir())
        elif status == 'failed':
            shutil.move(self.txpath.get_dir(), self.txpath.get_failed_dir())
        else:
            shutil.move(self.txpath.get_dir(), self.txpath.get_unknown_dir())

