#!/usr/bin/env python
#
# Copyright 2009 Marcelo Boveto Shima.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.    See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA    02110-1301, USA.
#
# Author: marceloshima@gmail.com (Marcelo Boveto Shima)

import os
import sys
import subprocess
import dbus
import signal
import threading
import getpass
import gobject
gobject.threads_init()

import tacix.nx.NXSession
import tacix.TXUtils
from tacix.nx.NXSession import NXSession
from tacix.nx import NXException

BUS_NAME="org.tacix"
INTERFACE_SESSION="org.tacix.Sessions"
INTERFACE_SERVER="org.tacix.Manager"
OBJECT_PATH="/org/tacix/Manager"

class NXNode:
    def __init__( self, logger=None, user=None ):
        self.dbus_signals = []
        self.txsession = None
        self.logger = logger
        self.txpath = None

        from dbus.mainloop.glib import DBusGMainLoop
        DBusGMainLoop(set_as_default=True)
        self.loop = gobject.MainLoop()

        self.bus = dbus.SystemBus()
        self.manager_obj = self.bus.get_object (BUS_NAME, OBJECT_PATH)
        self.manager = dbus.Interface (self.manager_obj, INTERFACE_SERVER)
        self.user = user
        if user is None:
            self.user = getpass.getuser()

    # Watchers
    # On session supended remove signals
    def watch_agent_info(self, info):
        if self.logger is not None:
            self.logger.info("Nxagent info is %s" % \
                info)

    def watch_agent_state_resume(self, state):
        return

    def watch_agent_state(self, state):
        if self.logger is not None:
            self.logger.info("Nxagent state is %s" % \
                state)

        if state in ['suspended', 'failed', 'terminated', 'denied']:
            self.loop.quit()

    # Helpers
    def send_signal(self, num):
        try:
            self.logger.info('Send signal %d to agent' % num)
            os.kill(int(self.txsession.GetProperty('txsession_pid')), num)
            return True
        except OSError, e:
            raise NXException('Could not send the signal to the agent pid')

    def get_property(self, property):
        return self.txsession.GetProperty(property)

    def add_state_signal(self, func):
        signal = self.txsession.connect_to_signal("StateChanged", func, dbus_interface=INTERFACE_SESSION)
        self.dbus_signals.append(signal)
        self.logger.debug('dbus_signals = %d' % len(self.dbus_signals))

    def add_info_signal(self, func):
        signal = self.txsession.connect_to_signal("InfoReceived", func, dbus_interface=INTERFACE_SESSION)
        self.dbus_signals.append(signal)
        self.logger.debug('dbus_signals = %d' % len(self.dbus_signals))

    def add_signals(self):
        self.add_state_signal(self.watch_agent_state)
        self.add_info_signal(self.watch_agent_info)

    def remove_signals(self):
        self.logger.debug('dbus_signals = %d' % len(self.dbus_signals))
        while len(self.dbus_signals) > 0:
            signal = self.dbus_signals.pop()
            signal.remove()
            self.logger.debug('dbus_signals = %d' % len(self.dbus_signals))
        self.logger.debug('Signals removed')


    # Actions functions
    def restore_session(self, options):
        self.get_session(options['id'])
        self.txsession.SetProperties( options )

        self.add_state_signal(self.watch_agent_state_resume)
        self.add_info_signal(self.watch_agent_info)
        self.send_signal(signal.SIGHUP)
        self.loop.run()

    def add_printer(self, options):
        self.txsession.AskPrinter(options)

    def attach_session(self, options):
        if 'id' in options:
            options['shadowed_id'] = options['id']
            del options['id']
        if 'display' in options:
            options['shadowdisplay'] = options['display']
            del options['display']
        self.start_session(options)

    def suspend_session(self, options):
        self.get_session(options['id'])
        self.send_signal(signal.SIGHUP)

    def terminate_session(self, options):
        self.get_session(options['id'])
        self.send_signal(signal.SIGTERM)

    def get_session(self, session_id):
        session_path = self.manager.GetSession (session_id)
        session_obj = self.bus.get_object (BUS_NAME, session_path)
        self.txsession = dbus.Interface (session_obj, INTERFACE_SESSION)

    def start_session(self, options):
        '''Init the NX Node'''
        # Create TXManager session
        session_path = self.manager.CreateSession (self.user)
        session_obj = self.bus.get_object (BUS_NAME, session_path)
        self.txsession = dbus.Interface (session_obj, INTERFACE_SESSION)
 
        self.txsession.SetProperties( options )

        session_id = self.get_property('id')
        os.environ['NXSESSIONID'] = session_id
        self.txpath = tacix.TXUtils.TXPath( session_id, create=True )
        if self.logger is None:
            self.logger = self.txpath.get_logger("nxnode")

        self.add_signals()
        self.txsession.Run()

    def wait_session(self):
        '''
        Wait for the session to quit
        '''
        self.logger.debug("Waiting for jobs to complete")
        self.loop.run()

