#!/usr/bin/env python
#
# Copyright 2009 Marcelo Boveto Shima.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.    See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA    02110-1301, USA.
#
# Author: marceloshima@gmail.com (Marcelo Boveto Shima)

import os
import random
import pwd

try:
    from hashlib import md5
except ImportError:
    from md5 import md5

import tacix.TXUtils
from tacix.nx import NXDisplay

allowed_options = {
                  # Enum values
                  'link' : ['modem', 'isdn', 'adsl', 'wan', 'lan'],
                  'mediahelper' : ['pulse', 'esd', 'artsd'],
                  'type' : ['unix-gnome', 'unix-kde', 'unix-cde', 'unix-xdm',
                      'unix-console', 'unix-default', 'unix-application',
                      'windows', 'vnc', 'shadow'],
                  'client' : ['linux'],
                  'clipboard' : ['both'],
                  'cache' : ['4M', '8M', '16M', '32M', '64M', '128M'], # In memory
                  'images' : ['0M', '32M', '64M', '128M', '256M', '512M'], # On disk

                  # Boolean values
                  'strict' : ['0', '1'],
                  'persistent' : ['1', '0'],
                  
                  # Client options
                  'backingstore' : ['1', '0'],
                  'encryption' : ['1', '0'],
                  'shmem' : ['1', '0'],
                  'shpix' : ['1', '0'], # Needs shmem enabled
                  'composite' : ['1', '0'],
                  'media' : ['0', '1'],
                  'render' : ['1', '0'],
                  'fullscreen' : ['0', '1'],
                  'virtualdesktop' : ['1', '0'],
                  'rootless' : ['0', '1'], # Needs virtualdesktop = 0, 0 = Disable X agent encoding
                  'taint' : ['1', '0'], # Needs rootless = 0, 0 = Disable taints of X replies
                  'stream' : ['1', '0'], # 0 = Disable zlib stream compression
                  'defer' : ['0', '1'], # 1 = Disable deferred screen updates

                  'application' : '',

                  # String client options
                  'imagecompressionmethod' : '', # None = ?, -1 = jpeg, 0 = None, 1 = jpeg + custom compression, 2 = rgb, 3 = jpeg + rgb, 4 = jpeg + rgb + custom compression
                  'imagecompressionlevel' : '', # Needs imagecompression 1 or 4, 0-9

                  # String values
                  'display' : '1000',
                  'session' : '',
                  'keyboard' : '',
                  'screeninfo' : '',
                  'geometry' : '', # 800x600

                  'txsession_pid' : '',

                  # Shadow properties
                  'shadowed_id' : '',
                  'shadowdisplay' : '',
                  'authority_file' : '',

                  # Windows and vnc properties
                  'agent_password' : '',
                  'agent_server' : '',
                  # Windows properties
                  'agent_domain' : '',
                  'agent_user' : ''
                  }

default_options = {
                  }

permissive = True

class NXOptions:
    def __init__( self, uid ):
        '''
           Create the options based on the default options
        '''
        self.private_options = {}
        self.options = {}

        if len(default_options) == 0:
            for key in allowed_options:
                value = allowed_options[key]
                if value is not None and type(value) is not str:
                    default_options[key] = value[0]

        self.options.update(default_options)
        self.calculate_options(uid)

    def get_id(self):
        return self.options['uniqueid']

    def get_cookie(self):
        return self.private_options['cookie']

    def get_display(self):
        return self.options['display']

    def get_state(self):
        return self.options['state']

    def set_state(self, state):
        old_state = self.options['state']
        if old_state != state:
            self.options['state'] = state
            return state
        return None

    def gen_cookie(self):
        return md5(str(random.random())).hexdigest()

    def get_allowed_value( self, property, value ):
        if permissive:
            return value
        new_value = None

        if property in allowed_options:
            avalues = allowed_options[property]
            if type(avalues) == str:
                new_value = avalues

            elif value not in avalues:
                new_value = avalues[0]

        if permissive:
            self.logger.debug('Permissive method allowed property %s with value %s instead of changing it to %s' %\
                    (property, value, new_value))
            return value

        self.logger.debug('Property %s with value %s not allowed, changing it to %s' %\
                (property, value, new_value))
        return new_value

    def calculate_options(self, uid):
        '''
           Calculate some options
        '''
        self.private_options['cookie'] = self.gen_cookie()
        self.private_options['proxy_cookie'] = self.private_options['cookie']

        display = str(NXDisplay.find_next_available_display(self.options['display'))
        self.options.update({
            'uniqueid' : self.gen_cookie().upper(),
            'display' : display,
            'hostname' : os.uname()[1],
            'state' : 'none',
            'cli_opts': '-RD--PS-', # FRD--PSA Used by listsessions
            'available': 'Yes', # N/A Used by listsessions
            'userip' : '127.0.0.1',
            'proxyip' : '127.0.0.1',
            'proxyip' : '127.0.0.1',
            'host' : '127.0.0.1',
            'uid' : uid,
            'user': pwd.getpwuid(int(uid)).pw_name
            })

        self.options['id'] = "%(hostname)s-%(display)s-%(uniqueid)s" % self.options

        for opt_name in self.options:
            self.options[opt_name] = self.get_allowed_value( opt_name, self.options[opt_name] )
        return

    def get_default_option( self, property ):
        if property in default_options:
            return default_options[property]
        return None

    def get_property( self, property ):
        if property in self.options:
            return self.options[property]
        return None

    def get_property_secure(self, property):
        if property in self.options:
            return self.options[property]

        if property in self.private_options:
            return self.private_options[property]
        return None

    def get_properties( self ):
        ret = {}
        ret.update(self.options)
        return ret

    def get_properties_secure( self ):
        ret = {}
        ret.update(self.options)
        ret.update(self.private_options)
        return ret

    def set_property( self, property, value ):
        handler_name = '_property_%s_handler' % property
        try:
            handler_method = getattr(self, handler_name)
            ret = handler_method(property, value)
            if not ret:
                return
        except AttributeError, e:
            pass

        ret = self.get_allowed_value(property, value)
        if ret == None:
            return

        self.options[property] = value

    def _property_screeninfo_handler(self, prop, value):
        split = value.split('x')
        self.options['depth'] = split[2].split('+')[0]
        self.options['screen'] = "%sx%s" % (split[0], split[1])
        return True

    def _property_mediahelper_handler(self, prop, value):
        if value == 'esd':
            self.options[prop] = 'pulse'
            return False
        return True

