#!/usr/bin/python
#
# Copyright 2009 Marcelo Boveto Shima.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.    See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA    02110-1301, USA.
#
# Author: marceloshima@gmail.com (Marcelo Boveto Shima)

import os
import signal
import subprocess
import threading
import traceback

from tacix.nx.NXAgentMonitor import NXAgentMonitor
from tacix.nx.NXSessionResources import NXSessionResources

BUS_NAME="org.tacix"
INTERFACE_SESSION="org.tacix.Sessions"
INTERFACE_SERVER="org.tacix.Manager"
OBJECT_PATH="/org/tacix/Manager"

class NXSession(NXSessionResources):
    def __init__ ( self, sessionid ):
        NXSessionResources.__init__(self, sessionid)

        signal.signal(signal.SIGHUP, self.signal_handler)
        signal.signal(signal.SIGTERM, self.signal_handler)

        self.nxagentmonitor = NXAgentMonitor(self)
        self.app_process = None
        self.app_signal = None

    # Watchers
    def signal_handler(self, num, stack):
        self.logger.info('Received signal %d' % num)
        if self.nxagent != None:
            self.logger.info('Sending signal to agent')
            self.nxagent.send_signal(num)

    # Agent actions
    def run( self ):
        os.environ['XDG_SESSION_COOKIE'] = self.txsession.OpenSession("")
        os.environ['NX_CLIENT'] = '/usr/NX/bin/nxclient'
        self.logger.debug("TXManager session openned\n")

        # Add watchers for the sessionapp
        if self.nxagent.options['type'] != 'shadow':
            self.setup_sound()
            if 'cups' in self.nxagent.options and self.nxagent.options['cups'] == '1':
                self.setup_printer_daemon()
            self.setup_signal_apps()
            self.setup_signal_resources()

        # Run the agent
        if not self.nxagent.run():
            return False
        self.nxagentmonitor.run()
        return True

    def wait( self ):
        self.loop.run()
        self.nxagent.send_signal(signal.SIGTERM)

        self.logger.debug("NXSession loop exited")
        self.nxagent.cleanup()


    # App actions
    def watch_agent_state_apps(self, state):
        self.logger.info("App watcher state is %s" % state)
        if state == 'running':
            self.app_signal.remove()
            self.run_app()

    def setup_signal_apps(self):
        self.logger.debug('Adding watcher for app')
        self.app_signal = self.add_dbus_signal(self.watch_agent_state_apps)

    def run_rdesktop(self):
        self.logger.info("NXSessionApp ready, launching rdesktop")
        log_file = open(self.txpath.get_filepath('logs/app.log'), 'w')

        if self.options['agent_domain'] != '':
            self.options['agent_domain'] = '-d ' + self.options['agent_domain']
        if self.options['application'] != '':
            self.options['application'] = '-s ' + self.options['application']
        cmd = 'rdesktop -u "%(agent_user)s" -k "%(agent_keyboard)s" %(agent_domain)s %(application)s -f %(agent_server)s' % self.options
        process = subprocess.Popen(cmd, shell=True, stdin=subprocess.PIPE, stdout=log_file, stderr=log_file, close_fds=True)
        process.stdin.write(self.options['agent_password'])
        # TODO Verify if the redesktop is not closed
        process.stdin.close()
        return process

    def run_vncviewer(self):
        self.logger.info("NXSessionApp ready, launching vncviewer")
        log_file = open(self.txpath.get_filepath('logs/app.log'), 'w')

        cmd = 'vnc' % self.options
        process = subprocess.Popen(cmd, shell=True, stdout=log_file, stderr=log_file, close_fds=True)
        return process

    def run_xsession(self):
        """Start the session application (startkde/gnome-session/etc)."""
        cmd = '/etc/tacix/Xsession'
        self.logger.info("NXSessionApp ready, launching app")
        log_file = open(self.txpath.get_filepath('logs/app.log'), 'w')

        os.environ['NX_SESSION_TYPE'] = self.nxagent.options['type']
        if 'application' in self.nxagent.options:
            os.environ['NX_APPLICATION'] = self.nxagent.options['application']
        return subprocess.Popen(cmd, stdout=log_file, stderr=log_file, close_fds=True)

    def run_app(self):
        type = self.nxagent.options['type']
        if type == 'windows':
            self.app_process = self.run_rdesktop()
        elif type == 'vnc':
            self.app_process = self.run_vncviewer()
        else:
            self.app_process = self.run_xsession()
        self.app_thread = threading.Thread(target=self.wait_app_process, name='app_thread')
        self.app_thread.start()

    def wait_app_process( self ):
        self.logger.debug("Waiting for nxapp")
        self.app_process.wait()
        self.nxagent.terminate()
        self.logger.debug("nxapp closed")

